@@ -1,7 +1,19 @@
+import logging
+
 import numpy as np
 
-from .typeutils import get_a_buffer, Info, is_integer, _to_slot_size
-from .scalar import Int64
+
+from .typeutils import (
+    get_a_buffer,
+    Info,
+    is_integer,
+    _to_slot_size,
+    default_conf,
+)
+from .scalar import Int64, is_scalar
+
+log = logging.getLogger(__name__)
+
 
 """
 array itemtype d1 d2 ...
@@ -38,6 +50,21 @@ Array instance:
 """
 
 
+def get_suffix(shape):
+    # write suffix  6x6x6 or Nx6xM
+    sshape = []
+    ilst = 0
+    lst = "NMOPQRSTUVWXYZABCDEFGHIJKLM"
+    # lst = "nmopqrstuvwxyzabcdefghijklm"
+    for dd in shape:
+        if dd is None:
+            sshape.append(lst[ilst])
+            ilst = (ilst + 1) % len(lst)
+        else:
+            sshape.append(str(dd))
+    return "x".join(sshape)
+
+
 def get_shape_from_array(value):
     if hasattr(value, "shape"):
         return value.shape
@@ -129,6 +156,11 @@ def bound_check(index, shape):
             raise IndexError(f"index {index} outside shape {shape}")
 
 
+class Index:
+    def __init__(self, cls):
+        self.cls = cls
+
+
 class MetaArray(type):
     def __new__(cls, name, bases, data):
         if "_itemtype" in data:  # specialized class
@@ -176,14 +208,39 @@ class MetaArray(type):
 
             data["_size"] = _size
             data["_data_offset"] = _data_offset
+        # need to applied to derived classes as well
+        if "_c_type" not in data:
+            data["_c_type"] = name
 
         return type.__new__(cls, name, bases, data)
 
     def __getitem__(cls, shape):
         return Array.mk_arrayclass(cls, shape)
 
+    def _get_offset(cls, index):
+        return get_offset(index, cls._strides)
+
+    def _get_n_items(cls):
+        if cls._is_static_shape:
+            return np.prod(cls._shape)
+        else:
+            raise ValueError("Cannot get n items from dynamic shapes")
+
+    def __repr__(cls):
+        return f"<array {cls.__name__}>"
+
 
 class Array(metaclass=MetaArray):
+    _shape: tuple
+    _order: tuple
+    _strides: tuple
+    _itemtype: type
+    _size: int
+    _dshape: tuple
+    _is_static_shape: bool
+    _is_static_type: bool
+    _data_offset: int
+
     @classmethod
     def mk_arrayclass(cls, itemtype, shape):
         if type(shape) in (int, slice):
@@ -198,21 +255,9 @@ class Array(metaclass=MetaArray):
             else:
                 nshape.append(dd)
 
-        if len(shape) <= 3:
-            lst = "NMOPQRSTU"
-            sshape = []
-            ilst = 0
-            for d in nshape:
-                if d is None:
-                    sshape.append(lst[ilst % len(lst)])
-                    ilst += 1
-                else:
-                    sshape.append(str(d))
+        suffix = get_suffix(nshape)
 
-            suffix = "by".join(sshape)
-        else:
-            suffix = f"{len(shape)}D"
-        name = itemtype.__name__ + "_" + suffix
+        name = f"Arr{suffix}{itemtype.__name__}"
 
         data = {
             "_itemtype": itemtype,
@@ -230,6 +275,7 @@ class Array(metaclass=MetaArray):
         - offsets
         - value: None if args contains dimensions else args[0]
         """
+        log.debug(f"get size for {cls} from {args}")
         info = Info()
         extra = {}
         if cls._size is not None:
@@ -242,7 +288,7 @@ class Array(metaclass=MetaArray):
                     raise ValueError(f"shape not valid for {args[0]} ")
                 value = args[0]
             elif len(args) > 1:
-                raise ValueError(f"too many arguments")
+                raise ValueError("too many arguments")
             size = cls._size
             shape = cls._shape
             order = cls._order
@@ -271,7 +317,7 @@ class Array(metaclass=MetaArray):
                         if ndim is None:
                             dshape.append(idim)
                         else:
-                            if shape[idem] != ndim:
+                            if shape[idim] != ndim:
                                 raise ValueError(
                                     "Array: incompatible dimensions"
                                 )
@@ -301,11 +347,13 @@ class Array(metaclass=MetaArray):
 
             # needs items, order, shape, value
             if cls._is_static_type:
-                offsets = np.empty(shape, dtype="int64")
-                for idx in iter_index(shape, order):
-                    offsets[idx] = offset
-                    offset += cls._itemtype._size
+                # offsets = np.empty(shape, dtype="int64")
+                # for idx in iter_index(shape, order):
+                #    offsets[idx] = offset
+                #    offset += cls._itemtype._size
+                offset += cls._itemtype._size * items
                 size = _to_slot_size(offset)
+
             else:
                 # args must be an array of correct dimensions
                 offsets = np.empty(shape, dtype="int64")
@@ -315,14 +363,15 @@ class Array(metaclass=MetaArray):
                     offsets[idx] = offset
                     offset += extra[idx].size
                 size = _to_slot_size(offset)
+                info.offsets = offsets
+                info.extra = extra
 
-        info.extra = extra
-        info.offsets = offsets
         info.shape = shape
         info.strides = strides
         info.size = size
         info.order = order
         info.value = value
+        info.items = np.prod(shape)
         return info
 
     @classmethod
@@ -346,7 +395,7 @@ class Array(metaclass=MetaArray):
             if len(self._shape) > 1:  # getting strides
                 # could be computed from shape and order but offset needs to taken
                 strides = []
-                for i in range(shape):
+                for _ in range(shape):
                     strides.append(Int64._from_buffer(self._buffer, coffset))
                     coffset += 8
             else:
@@ -358,7 +407,7 @@ class Array(metaclass=MetaArray):
         else:
             shape = cls._shape
         if not cls._is_static_type:
-            items = prod(shape)
+            items = np.prod(shape)
             self._offsets = Int64._array_from_buffer(buffer, coffset, items)
         return self
 
@@ -384,15 +433,45 @@ class Array(metaclass=MetaArray):
             coffset += 8 * len(header)
         if not cls._is_static_type:
             Int64._array_to_buffer(buffer, coffset, info.offsets)
-        if isinstance(value, np.ndarray) and hasattr(
-            cls._itemtype, "_dtype"
-        ):  # not robust try scalar classes
-            if cls._itemtype._dtype == value.dtype:
-                buffer.write(value.tobytes())
+            coffset += 8 * len(info.offsets)
+        if hasattr(cls._itemtype, "_dtype") and hasattr(
+            value, "dtype"
+        ):  # is a scalar type:
+            buffer.update_from_nplike(coffset, cls._itemtype._dtype, value)
+        elif isinstance(value, cls):
+            if value._size == info.size:
+                buffer.update_from_xbuffer(
+                    offset, value._buffer, value._offset, value._size
+                )
+            else:
+                raise ValueError("Value {value} not compatible size")
+        elif value is None:  # no value to initialize
+            if is_scalar(cls._itemtype):
+                pass  # leave uninitialized
+            else:
+                value = cls._itemtype()  # use default type
+                if cls._is_static_type:
+                    ioffset = offset + cls._data_offset
+                    for idx in range(info.items):
+                        cls._itemtype._to_buffer(buffer, ioffset, value, None)
+                        ioffset += cls._itemtype._size
+                else:
+                    for idx in iter_index(info.shape, cls._order):
+                        cls._itemtype._to_buffer(
+                            buffer,
+                            offset + info.offsets[idx],
+                            value[idx],
+                            info.extra.get(idx),
+                        )
+        else:  # there is a value for initialization
+            if cls._is_static_type:
+                ioffset = offset + cls._data_offset
+                for idx in iter_index(info.shape, cls._order):
+                    cls._itemtype._to_buffer(
+                        buffer, ioffset, value[idx], info=None
+                    )
+                    ioffset += cls._itemtype._size
             else:
-                buffer.write(value.astype(cls._itemtype._dtype).tobytes())
-        else:
-            if value is not None:
                 for idx in iter_index(info.shape, cls._order):
                     cls._itemtype._to_buffer(
                         buffer,
@@ -400,16 +479,6 @@ class Array(metaclass=MetaArray):
                         value[idx],
                         info.extra.get(idx),
                     )
-            else:
-                # TODO shortcuts for simple arrays
-                value = cls._itemtype()
-                for idx in iter_index(info.shape, cls._order):
-                    cls._itemtype._to_buffer(
-                        buffer,
-                        offset + info.offsets[idx],
-                        value,
-                        info.extra.get(idx),
-                    )
 
     def __init__(self, *args, _context=None, _buffer=None, _offset=None):
         # determin resources
@@ -437,10 +506,6 @@ class Array(metaclass=MetaArray):
         else:
             return self.__class__._size
 
-    @classmethod
-    def _get_offset(cls, index):
-        return get_offset(index, cls._strides)
-
     @classmethod
     def _get_position(cls, index):
         offset = get_offset(index, cls._strides)
@@ -482,48 +547,80 @@ class Array(metaclass=MetaArray):
                 )
             cls._itemtype._to_buffer(self._buffer, offset, value)
 
-    @classmethod
-    def _gen_method_specs(cls, base=None):
-        methods = []
-        if base is None:
-            base = []
-        spec = base + [cls]
-        methods.append(spec)
-        if hasattr(cls._itemtype, "_gen_method_specs"):
-            methods.extend(cls._itemtype._gen_method_specs(spec))
-        return methods
-
     def _get_offset(self, index):
-        return sum(ii * ss for ii, ss in zip(index, self._strides))
-
-    @classmethod
-    def _get_c_offset(cls, conf):
-        ctype = conf.get("ctype", "char")
-        itype = conf.get("itype", "int64_t")
-
-        out = []
-        if hasattr(cls, "_strides"):  # static shape or 1d dynamic shape
-            strides = cls._strides
+        if isinstance(index, (int, np.integer)):
+            index = (index,)
+        cls = self.__class__
+        if hasattr(self, "_offsets"):
+            offset = self._offset + self._offsets[index]
         else:
-            nd = len(cls._shape)
-            strides = []
-            stride_offset = 8 + nd * 8
-            for ii in range(nd):
-                sname = f"{itype} {cls.__name__}_s{ii}"
-                svalue = f"*({itype}*) (({ctype}*) obj+offset+{stride_offset})"
-                out.append(f"{sname}={svalue};")
-                strides.append(sname)
-
-        soffset = "+".join([f"i{ii}*{ss}" for ii, ss in enumerate(strides)])
-        if cls._data_offset > 0:
-            soffset = f"{cls._data_offset}+{soffset}"
-        if cls._is_static_type:
-            out.append(f"  offset+={soffset};")
-        else:
-            out.append(
-                f"  offset+=*({itype}*) (({ctype}*) obj+offset+{soffset});"
+            bound_check(index, self._shape)
+            offset = (
+                self._offset
+                + cls._data_offset
+                + get_offset(index, self._strides)
             )
-        return out
+        return offset
 
     def _iter_index(self):
         return iter_index(self._shape, self._order)
+
+    def __len__(self):
+        return np.prod(self._shape)
+
+    def to_nplike(self):
+        shape = self._shape
+        cshape = [shape[ii] for ii in self._order]
+        if hasattr(self._itemtype, "_dtype"):
+            arr = self._buffer.to_nplike(
+                self._offset + self._data_offset, self._itemtype._dtype, cshape
+            ).transpose(self._order)
+            assert arr.strides == self._strides
+            return arr
+        else:
+            raise NotImplementedError
+
+    @classmethod
+    def _gen_data_paths(cls, base=None):
+        paths = []
+        if base is None:
+            base = []
+        paths.append(base + [cls])
+        path = base + [cls, Index(cls)]
+        paths.append(path)
+        if hasattr(cls._itemtype, "_gen_data_paths"):
+            paths.extend(cls._itemtype._gen_data_paths(path))
+        return paths
+
+    @classmethod
+    def _gen_c_api(cls, conf=default_conf):
+        from . import capi
+        paths = cls._gen_data_paths()
+        return capi.gen_code(cls, paths, conf)
+
+    @classmethod
+    def _gen_c_decl(cls, conf=default_conf):
+        from . import capi
+        paths = cls._gen_data_paths()
+        return capi.gen_cdefs(cls, paths, conf)
+
+    @classmethod
+    def _gen_kernels(cls, conf=default_conf):
+        from . import capi
+        paths = cls._gen_data_paths()
+        return capi.gen_kernels(cls, paths, conf)
+
+    def __repr__(self):
+        return f"<{self.__class__.__name__}{self._shape} at {self._offset}>"
+
+    @classmethod
+    def _get_inner_types(cls):
+        return [cls._itemtype]
+
+
+def is_index(atype):
+    return isinstance(atype,Index)
+
+def is_array(atype):
+    return isinstance(atype,MetaArray)
+
