@@ -17,7 +17,6 @@ Layout:
 
 Current implementation:
     1) instance stores the actual offsets for dynamic offset. Although it is wasteful for memory, it avoids a double round trip. This hinges on the structure being frozen at initializations.
-    2) cache mechanism in place, but not used so far
 
 Struct class:
 - _size: class size, None if not static
@@ -28,7 +27,6 @@ Struct class:
 Struct instance:
 - _offsets: cached offsets of dynamic fields dict indexed by field.index
 - _sizes: cached sizes of dynamic fields dict indexed by field.index
-- _cache: cached python objects for __get__
 
 Field instance:
 - ftype
@@ -43,6 +41,8 @@ Field instance:
 
 """
 import logging
+from typing import Callable, Optional
+
 
 from .typeutils import (
     get_a_buffer,
@@ -50,11 +50,11 @@ from .typeutils import (
     Info,
     _to_slot_size,
     _is_dynamic,
+    default_conf,
 )
 
 from .scalar import Int64
 from .array import Array
-from . import capi
 
 
 log = logging.getLogger(__name__)
@@ -66,27 +66,24 @@ class Field:
     ):
         self.ftype = ftype
         self.default = default
-        if default_factory is not None:
-            self.get_default = default_factory
+        self.default_factory = default_factory
         self.index = None  # filled by class creation
         self.name = None  # filled by class creation
         self.offset = None  # filled by class creation
         self.is_reference = None  # filled by class creation
         self.has_update = False  # filled by class creation
         self.readonly = readonly
+        self.is_union = None  # filled by class creation
 
     def __repr__(self):
-        return f"<Field{self.index} {self.name} at {self.offset}>"
+        return f"<field{self.index} {self.name} at {self.offset}>"
 
     def __get__(self, instance, cls=None):
         if instance is None:
             return self
         else:
-            if self.index in instance._cache:
-                return instance._cache[self.index]
-            else:
-                offset = instance._offset + self.get_offset(instance)
-                return self.ftype._from_buffer(instance._buffer, offset)
+            ftype, offset = self.get_offset(instance)
+            return ftype._from_buffer(instance._buffer, offset)
 
     def __set__(self, instance, value):
         """
@@ -100,29 +97,39 @@ class Field:
 
         if hasattr(self.ftype, "_update"):
             self.__get__(instance)._update(value)
-        else:
-            offset = instance._offset + self.get_offset(instance)
-            self.ftype._to_buffer(instance._buffer, offset, value)
+        else:  # TODO check if below is really needed
+            ftype, offset = self.get_offset(instance)
+            ftype._to_buffer(instance._buffer, offset, value)
 
     def get_offset(self, instance):  # compatible with info
         if self.is_reference:
-            return instance._offsets[self.index]
+            reloffset = instance._offsets[self.index]
+            if self.is_union:
+                absoffset = instance._offset + reloffset
+                ftype = self.ftype._get_stored_type(
+                    instance._buffer, absoffset
+                )
+            else:
+                ftype = self.ftype
         else:
-            return self.offset
+            reloffset = self.offset
+            ftype = self.ftype
+        return ftype, instance._offset + reloffset
 
     def get_default(self):
-        if self.default is None:
-            return self.ftype()
+        if self.default_factory is None:
+            if self.default is None:
+                return self.ftype()
+            else:
+                return dispatch_arg(self.ftype, self.default)
         else:
-            return dispatch_arg(self.ftype, self.default)
+            return self.default_factory()
 
-    def _get_c_offset(self, conf):
-        itype = conf.get("itype", "int64_t")
-        if self.is_reference:
-            doffset = f"offset+{self.offset}"  # starts of data
-            return [f"  offset+=(({itype}*) obj)[{doffset}]"]
+    def value_from_args(self, arg):
+        if self.name in arg:
+            return arg[self.name]
         else:
-            return self.offset
+            return self.get_default()
 
 
 class MetaStruct(type):
@@ -164,8 +171,13 @@ class MetaStruct(type):
             def _get_size(self):
                 return self.__class__._size
 
-            def _inspect_args(cls, *args, **nargs):
-                return Info(size=cls._size, is_static_size=True)
+            def _inspect_args(cls, *args, **kwargs):
+                info = Info(size=cls._size, is_static_size=True)
+                if len(args) == 1:
+                    info.value = args[0]
+                else:
+                    info.value = kwargs
+                return info
 
         else:
             size = None
@@ -182,15 +194,17 @@ class MetaStruct(type):
                     offset += _to_slot_size(8)
                 # first dynamic field
                 d_fields[0].offset = offset
-                field.is_reference = False
+                d_fields[0].is_reference = False
 
             def _get_size(self):
                 return Int64._from_buffer(self._buffer, self._offset)
 
-            def _inspect_args(cls, *args, **nargs):
-                log.debug(f"get size for {cls} from {args} {nargs}")
-                if len(args) == 1:
+            def _inspect_args(cls, *args, **kwargs):
+                log.debug(f"get size for {cls} from {args} {kwargs}")
+                info = Info()
+                if len(args) == 1:  # is a dict or xobj
                     arg = args[0]
+                    info.value = arg
                     if isinstance(arg, dict):
                         offsets = {}  # offset of dynamic data
                         extra = {}
@@ -199,7 +213,7 @@ class MetaStruct(type):
                         ].offset  # offset first dynamic data
                         log.debug(f"{arg}")
                         for field in cls._d_fields:
-                            farg = arg.get(field.name, field.get_default())
+                            farg = field.value_from_args(arg)
                             log.debug(
                                 f"get size for field `{field.name}` using `{farg}`"
                             )
@@ -211,29 +225,46 @@ class MetaStruct(type):
                             offsets[field.index] = offset
                             offset += _to_slot_size(finfo.size)
                         # _offsets is used to because of field.get_offset(info)
-                        info = Info(size=offset, _offsets=offsets)
+                        info.size = offset
+                        info._offsets = offsets
                         if len(extra) > 0:
                             info.extra = extra
-                        return Info(size=offset, _offsets=offsets, extra=extra)
                     elif isinstance(arg, cls):
-                        size = arg._get_size()
-                        return Info(size=size)
+                        info.size = arg._get_size()
                     else:
                         raise ValueError(f"{arg} Not valid type for {cls}")
                 else:  # python argument
-                    return cls._inspect_args(nargs)
+                    return cls._inspect_args(kwargs)
+                return info
 
         data["_size"] = size
         data["_get_size"] = _get_size
         data["_inspect_args"] = classmethod(_inspect_args)
+        if "_c_type" not in data:
+            data["_c_type"] = name
 
         return type.__new__(cls, name, bases, data)
 
     def __getitem__(cls, shape):
         return Array.mk_arrayclass(cls, shape)
 
+    def __repr__(cls):
+        return f"<struct {cls.__name__}>"
+
 
 class Struct(metaclass=MetaStruct):
+    _fields: list
+    _d_fields: list
+    _inspect_args: Callable
+    _size: Optional[int]
+
+    @classmethod
+    def _pre_init(cls, *args, **kwargs):
+        return args, kwargs
+
+    def _post_init(self):
+        pass
+
     @classmethod
     def _from_buffer(cls, buffer, offset):
         self = object.__new__(cls)
@@ -245,50 +276,56 @@ class Struct(metaclass=MetaStruct):
             val = Int64._from_buffer(self._buffer, offset)
             _offsets[field.index] = val
         self._offsets = _offsets
-        self._cache = {}
+        self._post_init()
         return self
 
     @classmethod
     def _to_buffer(cls, buffer, offset, value, info=None):
         if isinstance(value, cls):  # binary copy
-            value_size = value._size
-            if value._buffer.context is buffer.context:
-                buffer.copy_from(
-                    value._buffer, value._offset, offset, value_size
-                )
-            else:
-                data = value._buffer.read(value._offset, value_size)
-                buffer.write(offset, data)
+            buffer.update_from_xbuffer(
+                offset, value._buffer, value._offset, value._size
+            )
         else:  # value must be a dict, again potential disctructive
-            if info is not None:
-                if hasattr(info, "_offsets"):
-                    cls._set_offsets(buffer, offset, info._offsets)
-                extra = getattr(info, "extra", {})
+            if info is None:
+                info = cls._inspect_args(value)
+            if cls._size is None:
+                Int64._to_buffer(buffer, offset, info.size)
+            if hasattr(
+                info, "_offsets"
+            ):  # if it has a least two dynamic fields
+                cls._set_offsets(buffer, offset, info._offsets)
+            extra = getattr(info, "extra", {})
             for field in cls._fields:
-                fvalue = value.get(field.name, field.get_default())
-                foffset = offset + field.get_offset(info)
+                fvalue = field.value_from_args(value)
+                if field.is_reference:
+                    foffset = offset + info._offsets[field.index]
+                else:
+                    foffset = offset + field.offset
                 finfo = extra.get(field.index)
                 field.ftype._to_buffer(buffer, foffset, fvalue, finfo)
 
     def _update(self, value):
         # check if direct copy is possible
         if isinstance(value, self.__class__) and value._size == self._size:
-            self._buffer.copy_from(
-                value._buffer, value._offset, self._offset, self._size
+            self._buffer.update_from_xbuffer(
+                self._offset, value._buffer, value._offset, value._size
             )
         else:
             for field in self._fields:
                 if field.name in value:
                     field.__set__(self, value[field.name])
 
-    def __init__(self, _context=None, _buffer=None, _offset=None, **nargs):
+    def __init__(
+        self, *args, _context=None, _buffer=None, _offset=None, **kwargs
+    ):
         """
         Create new struct in buffer from offset.
         If offset not provide
         """
         cls = self.__class__
+        args, kwargs = cls._pre_init(*args, **kwargs)
         # compute resources
-        info = cls._inspect_args(**nargs)
+        info = cls._inspect_args(*args, **kwargs)
         self._size = info.size
         # acquire buffer
         self._buffer, self._offset = get_a_buffer(
@@ -297,8 +334,8 @@ class Struct(metaclass=MetaStruct):
         # if dynamic struct store dynamic offsets
         if hasattr(info, "_offsets"):
             self._offsets = info._offsets  # struct offsets
-        self.__class__._to_buffer(self._buffer, self._offset, nargs, info)
-        self._cache = {}
+        cls._to_buffer(self._buffer, self._offset, info.value, info)
+        self._post_init()
 
     @classmethod
     def _set_offsets(cls, buffer, offset, loffsets):
@@ -340,31 +377,48 @@ class Struct(metaclass=MetaStruct):
             return longform
 
     @classmethod
-    def _gen_method_specs(cls, base=None):
-        methods = []
+    def _gen_data_paths(cls, base=None):
+        paths = []
         if base is None:
             base = []
+        paths.append(base + [cls])
         for field in cls._fields:
-            spec = base + [field]
-            methods.append(spec)
-            if hasattr(field.ftype, "_gen_method_specs"):
-                methods.extend(field.ftype._gen_method_specs(spec))
-        return methods
+            path = base + [cls, field]
+            paths.append(path)
+            if hasattr(field.ftype, "_gen_data_paths"):
+                paths.extend(field.ftype._gen_data_paths(path))
+        return paths
 
     @classmethod
-    def _gen_method_get_declaration(cls, conf={}):
-        specs_list = cls._gen_method_specs()
-        out = []
-        for specs in specs_list:
-            out.append(
-                capi.gen_method_get_declaration(cls.__name__, specs, conf)
-            )
-        return out
+    def _gen_c_api(cls, conf=default_conf):
+        from . import capi
+        paths = cls._gen_data_paths()
+        return capi.gen_code(cls, paths, conf)
 
     @classmethod
-    def _gen_method_get_body(cls, conf={}):
-        specs_list = cls._gen_method_specs()
-        out = []
-        for specs in specs_list:
-            out.append(capi.gen_method_get_body(cls.__name__, specs, conf))
-        return out
+    def _gen_c_decl(cls, conf=default_conf):
+        from . import capi
+        paths = cls._gen_data_paths()
+        return capi.gen_cdefs(cls, paths, conf)
+
+    @classmethod
+    def _gen_kernels(cls, conf=default_conf):
+        from . import capi
+        paths = cls._gen_data_paths()
+        return capi.gen_kernels(cls, paths, conf)
+
+    def _get_offset(self, fieldname):
+        for ff in self._fields:
+            if ff.name == fieldname:
+                return ff.get_offset(self)[1]
+
+    @classmethod
+    def _get_inner_types(cls):
+        return [fl.ftype for fl in cls._fields]
+
+def is_struct(atype):
+    return isinstance(atype,MetaStruct)
+
+def is_field(atype):
+    return isinstance(atype,Field)
+
